@@ -1,5 +1,5 @@
 /* java.lang.reflect.Field - reflection of Java fields
-   Copyright (C) 1998, 2001 Free Software Foundation, Inc.
+   Copyright (C) 1998, 2001, 2005 Free Software Foundation, Inc.
 
 This file is part of GNU Classpath.
 
@@ -15,8 +15,8 @@ General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with GNU Classpath; see the file COPYING.  If not, write to the
-Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
-02111-1307 USA.
+Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+02110-1301 USA.
 
 Linking this library statically or dynamically with other modules is
 making a combined work based on this library.  Thus, the terms and
@@ -43,6 +43,9 @@ This Classpath reference implementation has been modified to work with JamVM.
 
 package java.lang.reflect;
 
+import gnu.java.lang.ClassHelper;
+import gnu.java.lang.reflect.FieldSignatureParser;
+
 /**
  * The Field class represents a member variable of a class. It also allows
  * dynamic access to a member, via reflection. This works for both
@@ -114,6 +117,12 @@ extends AccessibleObject implements Member
     return name;
   }
 
+  /**
+   * Return the raw modifiers for this field.
+   * @return the field's modifiers
+   */
+  public native int getFieldModifiers(Class declaringClass, int slot);
+
   /**
    * Gets the modifiers this field uses.  Use the <code>Modifier</code>
    * class to interpret the values.  A field can only have a subset of the
@@ -123,11 +132,29 @@ extends AccessibleObject implements Member
    * @return an integer representing the modifiers to this Member
    * @see Modifier
    */
-  public int getModifiers() {
-      return getFieldModifiers(declaringClass, slot);
+  public int getModifiers()
+  {
+    return getFieldModifiers(declaringClass, slot);
   }
 
-  public native int getFieldModifiers(Class declaringClass, int slot);
+  /**
+   * Return true if this field is synthetic, false otherwise.
+   * @since 1.5
+   */
+  public boolean isSynthetic()
+  {
+    return (getFieldModifiers(declaringClass, slot) & Modifier.SYNTHETIC) != 0;
+  }
+
+  /**
+   * Return true if this field represents an enum constant,
+   * false otherwise.
+   * @since 1.5
+   */
+  public boolean isEnumConstant()
+  {
+    return (getFieldModifiers(declaringClass, slot) & Modifier.ENUM) != 0;
+  }
 
   /**
    * Gets the type of this field.
@@ -182,14 +209,24 @@ extends AccessibleObject implements Member
   public String toString()
   {
     // 64 is a reasonable buffer initial size for field
-    StringBuffer sb = new StringBuffer(64);
+    StringBuilder sb = new StringBuilder(64);
     Modifier.toString(getModifiers(), sb).append(' ');
-    sb.append(getType().getName()).append(' ');
+    sb.append(ClassHelper.getUserName(getType())).append(' ');
     sb.append(getDeclaringClass().getName()).append('.');
     sb.append(getName());
     return sb.toString();
   }
  
+  public String toGenericString()
+  {
+    StringBuilder sb = new StringBuilder(64);
+    Modifier.toString(getModifiers(), sb).append(' ');
+    sb.append(getGenericType()).append(' ');
+    sb.append(getDeclaringClass().getName()).append('.');
+    sb.append(getName());
+    return sb.toString();
+  }
+
   /**
    * Get the value of this Field.  If it is primitive, it will be wrapped
    * in the appropriate wrapper type (boolean = java.lang.Boolean).<p>
@@ -630,6 +667,31 @@ extends AccessibleObject implements Member
       setDField(o, declaringClass, type, slot, flag, 8, value);
   }
 
+  /**
+   * Return the generic type of the field. If the field type is not a generic
+   * type, the method returns the same as <code>getType()</code>.
+   *
+   * @throws GenericSignatureFormatError if the generic signature does
+   *         not conform to the format specified in the Virtual Machine
+   *         specification, version 3.
+   * @since 1.5
+   */
+  public Type getGenericType()
+  {
+    String signature = getSignature(declaringClass, slot);
+    if (signature == null)
+      return getType();
+    FieldSignatureParser p = new FieldSignatureParser(getDeclaringClass(),
+                                                      signature);
+    return p.getFieldType();
+  }
+
+  /**
+   * Return the String in the Signature attribute for this field. If there
+   * is no Signature attribute, return null.
+   */
+  private native String getSignature(Class declaringClass, int slot);
+
   private native double getDField(Object o, Class declaringClass, Class type, int slot, boolean noAccessCheck, int type_no);
   private native int getIField(Object o, Class declaringClass, Class type, int slot, boolean noAccessCheck, int type_no);
   private native long getJField(Object o, Class declaringClass, Class type, int slot, boolean noAccessCheck, int type_no);
