@@ -23,10 +23,13 @@
 extern void monitorInit(Monitor *mon);
 extern void monitorLock(Monitor *mon, Thread *self);
 extern void monitorUnlock(Monitor *mon, Thread *self);
-extern int monitorWait(Monitor *mon, Thread *self, long long ms, int ns);
+extern int monitorWait0(Monitor *mon, Thread *self, long long ms, int ns, int locked);
 extern int monitorNotify(Monitor *mon, Thread *self);
 extern int monitorNotifyAll(Monitor *mon, Thread *self);
 
+#define monitorWait(mon, self, ms, ns) \
+    monitorWait0(mon, self, ms, ns, FALSE)
+
 extern void objectLock(Object *ob);
 extern void objectUnlock(Object *ob);
 extern void objectNotify(Object *ob);
