@@ -25,11 +25,13 @@
 
 /* Thread states */
 
-#define CREATING     0
-#define STARTED      1
-#define RUNNING      2
-#define WAITING      3
-#define SUSPENDED    5
+#define CREATING      0
+#define STARTED       1
+#define RUNNING       2
+#define WAITING       3
+#define TIMED_WAITING 4
+#define BLOCKED       5
+#define SUSPENDED     6
 
 /* thread priorities */
 
@@ -46,30 +48,37 @@ typedef struct thread Thread;
 
 typedef struct monitor {
     pthread_mutex_t lock;
-    pthread_cond_t cv;
     Thread *owner;
-    int count;
-    int waiting;
-    int notifying;
-    int interrupting;
-    uintptr_t entering;
-    struct monitor *next;
     Object *obj;
+    int count;
+    int in_wait;
+    uintptr_t entering;
+    int wait_count;
+    Thread *wait_set;
+    struct monitor *next;
 } Monitor;
 
 struct thread {
+    int id;
+    pthread_t tid;
     char state;
-    char interrupted;
-    char interrupting;
     char suspend;
     char blocking;
-    pthread_t tid;
-    int id;
+    char interrupted;
+    char interrupting;
     ExecEnv *ee;
     void *stack_top;
     void *stack_base;
     Monitor *wait_mon;
+    Monitor *blocked_mon;
+    Thread *wait_prev;
+    Thread *wait_next;
+    pthread_cond_t wait_cv;
+    long long blocked_count;
+    long long waited_count;
     Thread *prev, *next;
+    unsigned int wait_id;
+    unsigned int notify_id;
 };
 
 extern Thread *threadSelf();
@@ -78,6 +87,11 @@ extern Thread *threadSelf0(Object *jThread);
 extern void *getStackTop(Thread *thread);
 extern void *getStackBase(Thread *thread);
 
+extern int getThreadsCount();
+extern int getPeakThreadsCount();
+extern void resetPeakThreadsCount();
+extern long long getTotalStartedThreadsCount();
+
 extern void threadInterrupt(Thread *thread);
 extern void threadSleep(Thread *thread, long long ms, int ns);
 extern void threadYield(Thread *thread);
@@ -96,6 +110,15 @@ extern void disableSuspend0(Thread *thread, void *stack_top);
 extern void enableSuspend(Thread *thread);
 extern void fastEnableSuspend(Thread *thread);
 
+extern Thread *attachJNIThread(char *name, char is_daemon, Object *group);
+extern void detachJNIThread(Thread *thread);
+
+extern char *getThreadStateString(Thread *thread);
+
+extern Thread *findThreadById(long long id);
+extern void suspendThread(Thread *thread);
+extern void resumeThread(Thread *thread);
+
 #define disableSuspend(thread)          \
 {                                       \
     sigjmp_buf *env;                    \
@@ -124,7 +147,7 @@ typedef pthread_mutex_t VMLock;
 }
 
 #define lockVMLock(lock, self) { \
-    self->state = WAITING;       \
+    self->state = BLOCKED;       \
     pthread_mutex_lock(&lock);   \
     self->state = RUNNING;       \
 }
@@ -153,7 +176,7 @@ typedef pthread_mutex_t VMLock;
         ts.tv_sec++;                                             \
         ts.tv_nsec -= 1000000000L;                               \
     }                                                            \
-    self->state = WAITING;                                       \
+    self->state = TIMED_WAITING;                                 \
     pthread_cond_timedwait(&wait_lock.cv, &wait_lock.lock, &ts); \
     self->state = RUNNING;                                       \
 }
